from base_st_classes import *
import threading

class MultipleSuite(NFSSuite):
    """Test multiple client interaction

    """
    def setUp(self):
        NFSSuite.setUp(self)
        self.fh1 = self.fh2 = self.fh3 = None
        self.state1 = self.state2 = self.state3 = None
        self.sync1 = threading.Event()
        self.sync2 = threading.Event()
        self.init_connection()
        self.init_connection(ncl=self.ncl2)

    def tearDown(self):
        self.ncl.close_file(self.fh1, self.state1, error=None)
        if self.state2:
            self.ncl2.close_file(self.fh2, self.state2, error=None)
        self.ncl3.close_file(self.fh3, self.state3, error=None)

        # restore permissions on dir to 0755
        stateid = stateid4(0, "")
        attrs = nfs4lib.dict2fattr({FATTR4_MODE:0755})
        ops = [self.ncl.putrootfh_op()] + self.ncl.lookup_path(self.tmp_dir)
        ops.append(self.ncl.setattr_op(stateid, attrs))
        self.ncl.do_ops(ops)
        
        try: self.clean_dir(self.tmp_dir)
        except SkipException, e:
            pass

    #
    # TESTS
    #
    def testTimedoutGrabLock(self):
        """MULTIPLE: server should release locks of timed out client 
        """
        # Client 1: create a file and get its fh
        attrs = nfs4lib.dict2fattr({FATTR4_MODE: 0770})
        self.fh1, self.state1 = self.ncl.create_confirm(attrs)

        # Client 2: open the file
        self.fh2, self.state2 = self.ncl2.open_confirm()

        # Client 1: lock file
        self.ncl.lock_file(self.fh1, self.state1)

        # Now wait, let client1 expire while client2 sends renews
        sleeptime = self.getLeaseTime() / 2
        for i in range(6):
            self.sleep(sleeptime)
            ops = [self.ncl2.renew_op(self.ncl2.clientid)]
            res = self.ncl2.do_ops(ops)
            self.assert_OK(res)

        # Client 2: Lock file, should work since Client 1's lock has expired
        self.ncl2.lock_file(self.fh2, self.state2, type=READ_LT)

    def testAccess(self):
        """MULTIPLE clients need access
        """
        # Client 1: create a file and get its fh
        attrs = nfs4lib.dict2fattr({FATTR4_MODE: 0777 })
        self.fh1, self.state1 = self.ncl.create_confirm(attrs)

        # Client 2: open the file
        self.fh2, self.state2 = self.ncl2.open_confirm()

    def testFailedOpen(self):
        """MULTIPLE: failed open should not mess up other clients' filehandles
        """

        # Set permissions on dir to exclude client 2 (tearDown will restore)
        stateid = stateid4(0, "")
        attrs = nfs4lib.dict2fattr({FATTR4_MODE:0700})
        lookup_dir_ops = self.ncl.lookup_path(self.tmp_dir)
        ops = [self.ncl.putrootfh_op()] + lookup_dir_ops
        ops.append(self.ncl.setattr_op(stateid, attrs))
        res = self.ncl.do_ops(ops)
        self.assert_OK(res)

        # Client 1: create a file and get its fh
        self.fh1, self.state1 = self.ncl.create_confirm()

        # Client 2: try to open the file
        self.ncl2.open_confirm(error=[NFS4ERR_ACCESS])
        
        # Client 1: recreate the file and get its fh
        self.fh1, self.state1 = self.ncl.create_confirm()

    def testGrabLock(self):
        """MULTIPLE clients trying to get lock
        """
        # Client 1: create a file and get its fh
        attrs = nfs4lib.dict2fattr({FATTR4_MODE: 0770})
        self.fh1, self.state1 = self.ncl.create_confirm(attrs)

        # Client 2: open the file
        self.fh2, self.state2 = self.ncl2.open_confirm()

        # Client 1: lock file
        lockid1 = self.ncl.lock_file(self.fh1, self.state1)

        # Client 2: Lock file, should not work since Client 1 has lock
        self.ncl2.lock_file(self.fh2, self.state2, type=READ_LT,
                            error=[NFS4ERR_DENIED])

        # Client 1: unlock file
        self.ncl.unlock_file(self.fh1, lockid1, 1)

        # Client 2: Lock file, should work now
        self.ncl2.lock_file(self.fh2, self.state2, type=READ_LT)

        
    def testReadLocks(self):
        """MULTIPLE clients all trying to get read locks"""
        # Client 1: create a file and get its fh
        attrs = nfs4lib.dict2fattr({FATTR4_MODE: 0770})
        self.fh1, self.state1 = self.ncl.create_confirm(attrs)

        # Client 2: open the file
        self.fh2, self.state2 = self.ncl2.open_confirm()

        # Client 1: readlock file 
        lockid1 = self.ncl.lock_file(self.fh1, self.state1, type=READ_LT)

        # Client 2: also readlock file
        lockid2 = self.ncl2.lock_file(self.fh2, self.state2, type=READ_LT)

        # Client 1: writelock file - should fail
        self.ncl.relock_file(self.fh1, lockid1, 1, type=WRITE_LT,
                             error=[NFS4ERR_DENIED])

        # Client 2: writelock file - should fail
        self.ncl2.relock_file(self.fh2, lockid2, 1, type=WRITE_LT,
                              error=[NFS4ERR_DENIED])
        
    def testWrongData(self):
        """MULTIPLE clients using other client's stateid...seems should return NFS4ERR_BAD_STATEID"""
        # Client 1: create a file and get its fh
        attrs = nfs4lib.dict2fattr({FATTR4_MODE: 0770})
        self.fh1, self.state1 = self.ncl.create_confirm(attrs)

        # Client 2: open the file
        self.fh2, self.state2 = self.ncl2.open_confirm()

        # Client 1: lock file using client2's stateid
        lockid = self.ncl.lock_file(self.fh1, self.state2,
                                    seqid=self.ncl2.get_seqid(advance=False),
                                    error=[NFS4ERR_BAD_STATEID])

    def _threadSample(self):
        """ thread function to implement competing clients
        """
        print "Entering thread"
        # Client 2: open the file
        self.fh2, self.state2 = self.ncl2.open_confirm()

        # Client 2: also read lock file
        lockid2 = self.ncl2.lock_file(self.fh2, self.state2, type=READ_LT)

        # Client 2: upgrade end of file to a blocking write
        self.ncl2.relock_file(self.fh2, lockid2, 1, 100, 50, type=WRITEW_LT,
                              error=[NFS4ERR_DENIED])

        # Notify main thread we are done, then wait for it
        print "Finished thread locks"
        self.sync1.set()
        print "Leaving thread"

    def XXXtestDeadlock(self):
        """MULTIPLE clients trying to deadlock

        This is mainly a sample of a way to run multiple threads.
        Currently, CITI server does not distinguish between normal and
        blocking locks, so this won't deadlock.
        """
        print
        print "Entering Deadlock"
        # Set up threaded to compete with this one
        self.sync1.clear()
        self.sync2.clear()
        thread = threading.Thread(target=self._threadSample)
        
        # Client 1: create a file and get its fh
        attrs = nfs4lib.dict2fattr({FATTR4_MODE: 0770})
        self.fh1, self.state1 = self.ncl.create_confirm(attrs)

        # Client 1: read lock file 
        lockid1 = self.ncl.lock_file(self.fh1, self.state1, type=READ_LT)

        print "Telling thread to start"
        # Wait for Client 2 to finish getting locks
        thread.start()
        self.sync1.wait()
        self.sync1.clear()

        print "Main starting up again"
        # Client 1: upgrade beginning of file to a blocking write
        self.ncl.relock_file(self.fh1, lockid1, 1, 0, 50, type=WRITEW_LT,
                             error=[NFS4ERR_DENIED])
        time.sleep(1)
        print "Leaving Deadlock"
        
    def testFairness(self):
        """MULTIPLE blocking locks may or may not be fair

        Since CITI server does not implement blocking locks, this should return
        that locking is unfair
        """

        # NOTE - there is a problem with using client 3.  If run as a single
        # test this works fine.  If run in conjuction with other tests,
        # especially those that sleep for significant periods of time,
        # the tcp connection for ncl3 can time out before a test is run
        # which uses it.

        # Client 1: create a file and get its fh

        attrs = nfs4lib.dict2fattr({FATTR4_MODE: 0770})
        self.fh1, self.state1 = self.ncl.create_confirm(attrs)

        # Client 1: lock file
        lockid1 = self.ncl.lock_file(self.fh1, self.state1)

        # Client 2: open the file
        self.fh2, self.state2 = self.ncl2.open_confirm()
        
        # Client 2: Lock file, should not work since Client 1 has lock
        self.ncl2.lock_file(self.fh2, self.state2, type=WRITEW_LT,
                            error=[NFS4ERR_DENIED])

        # Client 1: unlock file
        self.ncl.unlock_file(self.fh1, lockid1, 1)

        # Client 3: open the file
        self.init_connection(ncl=self.ncl3)
        self.fh3, self.state3 = self.ncl3.open_confirm()
        
        # Client 3: Lock file, may not work since Client 2 has pending lock
        lockid3 = self.ncl3.lock_file(self.fh3, self.state3, type=WRITEW_LT,
                                error=[NFS4ERR_DENIED,NFS4_OK])
        if lockid3:
            self.info_message("Locking is not fair")
            return

        # Client 2: Lock file, should work
        self.ncl2.lock_file(self.fh2, self.state2, type=WRITEW_LT)

        
    def testMultiport(self):
        """MULTIPLE clients on multiple ports

        This test passes, but sporadicaly in future init_connections...
        Causes :   File "./rpc.py", line 441, in sendfrag
               :     sock.send(header + frag)
               : error: (32, 'Broken pipe')
        or
        Causes : AssertionError: EOFError: short response

        Further investigation shows this happens when the server
        sends a TCP reset.
        """
        self.fail("This passes, but sporadically causes other init_connections to fail")
        # Create lots of clients, have one of them open file
        for i in range(150):
            ncl = self.create_client(0, 0, "garbage")
        self.init_connection(ncl=ncl)
        ncl.open_confirm(self.regfile[:-1],self.regfile[-1])

        # Try to get it to fail in this test
        # MultiClientSuite.testReadLocks(self)

    def testLotsOfLocks(self):
        """MULTIPLE clients with multiple locks"""
        
        # Client 1: create a file and get its fh
        attrs = nfs4lib.dict2fattr({FATTR4_MODE: 0770})
        self.fh1, self.state1 = self.ncl.create_confirm(attrs)

        # Client 1: lock file
        lockid1 = self.ncl.lock_file(self.fh1, self.state1, 0, 10, WRITE_LT)
        lockid1 = self.ncl.relock_file(self.fh1, lockid1, 1, 20, 20, READ_LT)

        # Client 2: open the file
        self.fh2, self.state2 = self.ncl2.open_confirm()

        # Client 2: lock file
        lockid2 = self.ncl2.lock_file(self.fh2, self.state2, 10, 10, WRITE_LT)
        lockid2 = self.ncl2.relock_file(self.fh2, lockid2, 1, 30, 20, READ_LT)

        # Client 3: open the file
        self.init_connection(ncl=self.ncl3)
        self.fh3, self.state3 = self.ncl3.open_confirm()

        # Client 3: test locks
        fh = self.fh3
        self.ncl3.lock_test(fh,  0, 10,  READ_LT, error=[NFS4ERR_DENIED])
        self.ncl3.lock_test(fh,  0, 10, WRITE_LT, error=[NFS4ERR_DENIED])
        self.ncl3.lock_test(fh, 10, 10,  READ_LT, error=[NFS4ERR_DENIED])
        self.ncl3.lock_test(fh, 10, 10, WRITE_LT, error=[NFS4ERR_DENIED])
        self.ncl3.lock_test(fh, 20, 30,  READ_LT, error=[NFS4_OK])
        self.ncl3.lock_test(fh, 20, 30, WRITE_LT, error=[NFS4ERR_DENIED])
        self.ncl3.lock_test(fh, 50, 10,  READ_LT, error=[NFS4_OK])
        self.ncl3.lock_test(fh, 50, 10, WRITE_LT, error=[NFS4_OK])

        # Client 1: upgrade locks
        fh = self.fh1
        self.ncl.relock_file(fh, lockid1, 2, 10, 10, READ_LT,
                             error=[NFS4ERR_DENIED])
        self.ncl.relock_file(fh, lockid1, 3, 10, 10, WRITE_LT,
                             error=[NFS4ERR_DENIED])
        lockid1 = self.ncl.relock_file(fh, lockid1, 4, 20, 10, WRITE_LT,
                                       error=[NFS4_OK])
        self.ncl.relock_file(fh, lockid1, 5, 30, 10, WRITE_LT,
                             error=[NFS4ERR_DENIED])
        self.ncl.relock_file(fh, lockid1, 6, 40, 10, WRITE_LT,
                             error=[NFS4ERR_DENIED])
        lockid1 = self.ncl.relock_file(fh, lockid1, 7, 40, 10, READ_LT,
                                       error=[NFS4_OK])
        lockid1 = self.ncl.relock_file(fh, lockid1, 8, 50, 10, WRITE_LT,
                                       error=[NFS4_OK])

        # Client 1: restore locks to original
        lockid1 = self.ncl.relock_file(fh, lockid1, 9, 20, 10, READ_LT,
                                       error=[NFS4_OK])
        lockid1 = self.ncl.unlock_file(fh, lockid1, 10, 40, 20)

        # Client 2: upgrade locks
        fh = self.fh2
        self.ncl2.relock_file(fh, lockid2, 2, 0, 10, READ_LT,
                              error=[NFS4ERR_DENIED])
        self.ncl2.relock_file(fh, lockid2, 3, 0, 10,WRITE_LT,
                              error=[NFS4ERR_DENIED])
        self.ncl2.relock_file(fh, lockid2, 4, 20, 10, WRITE_LT,
                              error=[NFS4ERR_DENIED])
        lockid2 = self.ncl2.relock_file(fh, lockid2, 5, 20, 10, READ_LT,
                                        error=[NFS4_OK])
        self.ncl2.relock_file(fh, lockid2, 6, 30, 10, WRITE_LT,
                              error=[NFS4ERR_DENIED])
        lockid2 = self.ncl2.relock_file(fh, lockid2, 7, 40, 10, WRITE_LT,
                                        error=[NFS4_OK])
        lockid2 = self.ncl2.relock_file(fh, lockid2, 8, 50, 10, WRITE_LT,
                                        error=[NFS4_OK])

        # Client 2: restore locks to original
        lockid2 = self.ncl2.relock_file(fh, lockid2,  9, 40, 10, READ_LT,
                                        error=[NFS4_OK])
        lockid2 = self.ncl2.unlock_file(fh, lockid2, 10, 20, 10)
        lockid2 = self.ncl2.unlock_file(fh, lockid2, 11, 50, 10)

        # Client 3: test locks
        fh = self.fh3
        self.ncl3.lock_test(fh,  0, 10,  READ_LT, error=[NFS4ERR_DENIED])
        self.ncl3.lock_test(fh,  0, 10, WRITE_LT, error=[NFS4ERR_DENIED])
        self.ncl3.lock_test(fh, 10, 10,  READ_LT, error=[NFS4ERR_DENIED])
        self.ncl3.lock_test(fh, 10, 10, WRITE_LT, error=[NFS4ERR_DENIED])
        self.ncl3.lock_test(fh, 20, 30,  READ_LT, error=[NFS4_OK])
        self.ncl3.lock_test(fh, 20, 30, WRITE_LT, error=[NFS4ERR_DENIED])
        self.ncl3.lock_test(fh, 50, 10,  READ_LT, error=[NFS4_OK])
        self.ncl3.lock_test(fh, 50, 10, WRITE_LT, error=[NFS4_OK])

    def testReadLock(self):
        """MULTIPLE read locked file"""
        # Client 1: create a file and get its fh
        attrs = nfs4lib.dict2fattr({FATTR4_MODE: 0770})
        self.fh1, self.state1 = self.ncl.create_confirm(attrs)

        # Client 1: writelock file
        lockid1 = self.ncl.lock_file(self.fh1, self.state1)

        # Client 2: open the file
        self.fh2, self.state2 = self.ncl2.open_confirm()

    def testDenyRead(self):
        """MULTIPLE read a read-denied file"""
        # Client1: create a file with deny-read
        attrs = nfs4lib.dict2fattr({FATTR4_MODE: 0770})
        self.fh1, self.state1 = self.ncl.create_confirm(attrs, deny=OPEN4_SHARE_DENY_READ)

        # Client 2: open the file for read access should fail
        self.ncl2.open_confirm(error=[NFS4ERR_SHARE_DENIED])

        # Client 2: try to read file without opening
        self.ncl2.read_file(error=[NFS4ERR_SHARE_DENIED])

        # Client 2: try to write file without opening
        self.ncl2.write_file()

        # Client 2: open the file for write access
        self.ncl2.open_confirm(access=OPEN4_SHARE_ACCESS_WRITE)

    def testDenyWrite(self):
        """MULTIPLE read a write-denied file"""
        # Client1: create a file with deny-write
        attrs = nfs4lib.dict2fattr({FATTR4_MODE: 0770})
        self.fh1, self.state1 = self.ncl.create_confirm(attrs, deny=OPEN4_SHARE_DENY_WRITE)

        # Client 2: open the file for write access should fail
        self.ncl2.open_confirm(access=OPEN4_SHARE_ACCESS_WRITE,
                               error=[NFS4ERR_SHARE_DENIED])

        # Client 2: try to write file without opening
        self.ncl2.write_file(error=[NFS4ERR_SHARE_DENIED])

        # Client 2: try to read file without opening
        self.ncl2.read_file()

        # Client 2: open the file for reading
        self.ncl2.open_confirm()

    def testDenyBoth(self):
        """MULTIPLE read and write a read-write-denied file"""
        # Client1: create a file with deny-both
        attrs = nfs4lib.dict2fattr({FATTR4_MODE: 0770})
        self.fh1, self.state1 = self.ncl.create_confirm(attrs, deny=OPEN4_SHARE_DENY_BOTH)

        # Client 2: open the file for read access should fail
        self.ncl2.open_confirm(error=[NFS4ERR_SHARE_DENIED])

        # Client 2: open the file for write access should fail
        self.ncl2.open_confirm(access=OPEN4_SHARE_ACCESS_WRITE,
                               error=[NFS4ERR_SHARE_DENIED])

        # Client 2: try to read file without opening
        self.ncl2.read_file(error=[NFS4ERR_SHARE_DENIED])

        # Client 2: try to write file without opening
        self.ncl2.write_file(error=[NFS4ERR_SHARE_DENIED])

